from multiprocessing.dummy import Pool, Queue

from lemoncheesecake.exceptions import TaskFailure, TasksExecutionFailure, CircularDependencyError, \
    serialize_current_exception

_KEYBOARD_INTERRUPT_ERROR_MESSAGE = "all tests have been interrupted by the user"


class BaseTask(object):
    def __init__(self):
        self.successful = None
        self.exception = None

    def get_all_dependencies(self):
        return self.get_on_completion_dependencies() + self.get_on_success_dependencies()

    def get_on_success_dependencies(self):
        return []

    def get_on_completion_dependencies(self):
        return []

    def run(self, context):
        pass

    def skip(self, context, reason):
        pass


def pop_runnable_tasks(remaining_tasks, completed_tasks, nb_tasks):
    runnable_tasks = [
        task for task in remaining_tasks if set(task.get_all_dependencies()).issubset(completed_tasks)
    ]
    # return the tasks with the greater number of dependencies first
    runnable_tasks_by_priority = sorted(
        runnable_tasks, key=lambda task: len(task.get_all_dependencies()), reverse=True
    )

    for task in runnable_tasks_by_priority[:nb_tasks]:
        remaining_tasks.remove(task)
        yield task


def pop_skippable_tasks(ref_task, tasks):
    for task in list(tasks):
        if ref_task in task.get_on_success_dependencies() and task in tasks:
            tasks.remove(task)
            yield task
            for indirect_task in pop_skippable_tasks(task, tasks):
                yield indirect_task


def run_task(task, context, completed_task_queue):
    try:
        task.run(context)
    except TaskFailure:
        task.successful = False
    except Exception:
        task.successful = False
        task.exception = serialize_current_exception()
    else:
        task.successful = True

    completed_task_queue.put(task)


def handle_task(task, watchdogs, context, completed_task_queue):
    for watchdog in watchdogs:
        error = watchdog(task)
        if error:
            skip_task(task, context, completed_task_queue, reason=error)
            return

    run_task(task, context, completed_task_queue)


def schedule_tasks_to_be_run(tasks, watchdogs, context, pool, completed_tasks_queue):
    for task in tasks:
        pool.apply_async(handle_task, args=(task, watchdogs, context, completed_tasks_queue))


def skip_task(task, context, completed_task_queue, reason=""):
    try:
        task.skip(context, reason)
    except Exception:
        task.successful = False
        task.exception = serialize_current_exception()
    else:
        task.successful = True

    completed_task_queue.put(task)


def schedule_tasks_to_be_skipped(tasks, context, pool, completed_tasks_queue, reason=""):
    for task in tasks:
        pool.apply_async(skip_task, args=(task, context, completed_tasks_queue, reason))


def skip_all_tasks(tasks, remaining_tasks, completed_tasks, context, pool, completed_tasks_queue, reason):
    schedule_tasks_to_be_skipped(remaining_tasks, context, pool, completed_tasks_queue, reason)
    while len(completed_tasks) != len(tasks):
        completed_task = completed_tasks_queue.get()
        completed_tasks.append(completed_task)


def run_tasks(tasks, context=None, nb_threads=1, watchdog=None):
    got_keyboard_interrupt = False
    watchdogs = [lambda _: _KEYBOARD_INTERRUPT_ERROR_MESSAGE if got_keyboard_interrupt else None]
    if watchdog:
        watchdogs.append(watchdog)

    for task in tasks:
        check_task_dependencies(task)

    remaining_tasks = list(tasks)
    completed_tasks = list()

    pool = Pool(nb_threads)
    completed_tasks_queue = Queue()

    try:
        schedule_tasks_to_be_run(
            pop_runnable_tasks(remaining_tasks, completed_tasks, nb_threads),
            watchdogs, context, pool, completed_tasks_queue
        )

        while len(completed_tasks) != len(tasks):
            # wait for one task to complete
            completed_task = completed_tasks_queue.get()
            completed_tasks.append(completed_task)

            # schedule tasks to be skipped after task failure
            if not completed_task.successful:
                tasks_to_be_skipped = pop_skippable_tasks(completed_task, remaining_tasks)
                schedule_tasks_to_be_skipped(tasks_to_be_skipped, context, pool, completed_tasks_queue)

            # schedule tasks to be run waiting for task success or simple completion
            tasks_to_be_run = pop_runnable_tasks(remaining_tasks, completed_tasks, nb_threads)
            schedule_tasks_to_be_run(tasks_to_be_run, watchdogs, context, pool, completed_tasks_queue)

    except KeyboardInterrupt:
        got_keyboard_interrupt = True
        skip_all_tasks(
            tasks, remaining_tasks, completed_tasks, context, pool, completed_tasks_queue,
            _KEYBOARD_INTERRUPT_ERROR_MESSAGE
        )

    finally:
        pool.close()

    if any(task.exception for task in tasks):
        raise TasksExecutionFailure(
            "Caught exceptions:\n%s" % "\n".join(task.exception for task in tasks if task.exception)
        )


def check_task_dependencies(task, task_path=()):
    for task_in_path in task_path:
        if task_in_path in task.get_all_dependencies():
            raise CircularDependencyError("Task %s has a circular dependency on task %s" % (task, task_in_path))

    for dependency in task.get_all_dependencies():
        check_task_dependencies(dependency, task_path=(task,) + task_path)



import re
import time

from lemoncheesecake.events import TestSessionSetupEndEvent, TestSessionTeardownEndEvent, \
    TestEndEvent, SuiteSetupEndEvent, SuiteTeardownEndEvent, SuiteEndEvent, SteppedEvent


def _get_testish_info(event, report):
    if isinstance(event, TestEndEvent):
        test_data = report.get_test(event.test)
        return True, test_data.status == "passed"

    if isinstance(event, SuiteSetupEndEvent):
        suite_data = report.get_suite(event.suite)
        return True, suite_data.suite_setup and not suite_data.suite_setup.is_successful()

    if isinstance(event, SuiteTeardownEndEvent):
        suite_data = report.get_suite(event.suite)
        return True, suite_data.suite_teardown and not suite_data.suite_teardown.is_successful()

    if isinstance(event, TestSessionSetupEndEvent):
        return True, report.test_session_setup and not report.test_session_setup.is_successful()

    if isinstance(event, TestSessionTeardownEndEvent):
        return True, report.test_session_teardown and not report.test_session_teardown.is_successful()

    return False, None


def save_at_each_suite_strategy(event, _):
    return isinstance(event, SuiteEndEvent)


def save_at_each_test_strategy(event, report):
    is_testish_end, _ = _get_testish_info(event, report)
    return is_testish_end


def save_at_each_failed_test_strategy(event, report):
    is_testish_end, is_successful = _get_testish_info(event, report)
    return is_testish_end and not is_successful


def save_at_each_event_strategy(event, _):
    # "each_event" means events that produces actual data, meaning events that adds data in
    # a step
    return isinstance(event, SteppedEvent)


class SaveAtInterval(object):
    def __init__(self, interval):
        self.interval = interval
        self.last_saving = None

    def __call__(self, event, report):
        now = time.time()
        if self.last_saving:
            must_be_saved = now > self.last_saving + self.interval
            if must_be_saved:
                self.last_saving = now
            return must_be_saved
        else:
            self.last_saving = now  # not a saving but an initialization
            return False


def make_report_saving_strategy(expression):
    # first, try with a static expression
    static_expressions = {
        "at_end_of_tests": None,  # no need to an intermediate report saving in this case
        "at_each_suite": save_at_each_suite_strategy,
        "at_each_test": save_at_each_test_strategy,
        "at_each_failed_test": save_at_each_failed_test_strategy,
        "at_each_event": save_at_each_event_strategy
    }

    try:
        return static_expressions[expression]
    except KeyError:
        pass

    # second, try with "every_Ns"
    m = re.compile("^every[_ ](\d+)s$").match(expression)
    if m:
        return SaveAtInterval(int(m.group(1)))

    # ok... nothing we know about
    raise ValueError("Unknown report saving strategy expression '%s'" % expression)
